@@ -1,12 +1,19 @@
 # -*- coding:utf-8 -*-
 from __future__ import unicode_literals
 from future.builtins import object
+from collections import namedtuple
+
 
 from limpyd.utils import unique_key
 from limpyd.exceptions import *
 from limpyd.fields import MultiValuesField
 
 
+ParsedFilter = namedtuple('ParsedFilter', ['index', 'suffix', 'extra_field_parts', 'value'])
+
+NONE_SLICE = slice(None, None, None)
+
+
 class CollectionManager(object):
     """
     Retrieve objects collection, optionnaly slice and order it.
@@ -23,10 +30,12 @@ class CollectionManager(object):
     Slicing a collection will force a sort.
     """
 
+    _accepted_key_types = {'set'}  # Type of keys indexes are allowed to return
+
     def __init__(self, cls):
         self.cls = cls
         self._lazy_collection = {  # Store infos to make the requested collection
-            'sets': set(),  # store sets to use (we'll intersect them)
+            'sets': [],  # store sets to use (we'll intersect them)
             'pks': set(),  # store special filter on pk
         }
         self._instances = False  # True when instances are asked
@@ -34,7 +43,7 @@ class CollectionManager(object):
         self._instances_skip_exist_test = False  # If True will return instances
                                                  # without testing if pk exist
         self._sort = None  # Will store sorting parameters
-        self._slice = None  # Will store slice parameters (start and num)
+        self._sort_limits = None  # Will store slice parameters (start and num)
         self._len = None  # Store the result of the final collection, to avoid
                           # having to compute the whole thing twice when doing
                           # `list(Model.collection)` (a `list` will call
@@ -44,45 +53,142 @@ class CollectionManager(object):
                                 # True by default to manage the __iter__ + __len__
                                 # case, specifically set to False in other cases
 
+        self.scripted_slicing = True
+
     def __iter__(self):
-        self._len_mode = False
-        return self._collection.__iter__()
+        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
+        try:
+            self._len_mode = False
+            return self._collection.__iter__()
+        finally:
+            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode
+
+    def _optimize_slice(self, the_slice, can_reverse):
+        """
+
+        Parameters
+        ----------
+        the_slice: slice
+            The python slice to optimize
+        can_reverse: bool
+            If we can ask redis to reverse the data before slicing.
+
+        Returns
+        -------
+        tuple
+            Four elements:
+            - optimized: bool
+                If the slice is optimized
+            - start: int or None
+                start item to pass to the redis command
+                If None, we want from the very start
+            - count: int or None
+                number of items to ask the redi command
+                If None, we want everything from the start
+            - rev: bool
+                If we should ask redis to work the reverse way
+            - python_slice
+                The slice to be applied on the python side
+
+        """
+        step = 1 if the_slice.step in (1, None) else the_slice.step
+        start = (the_slice.start or 0) if step > 0 else the_slice.start
+        stop = the_slice.stop
+
+        if start is not None and stop is not None:
+
+            # cases where we know the result is always empty
+            if step > 0 and (0 <= stop <= start or stop <= start < 0):
+                return True, None, None, False, NONE_SLICE
+
+            if step < 0 and (0 <= start <= stop or start <= stop < 0):
+                return True, None, None, False, NONE_SLICE
+
+            # if start and stop are on the same side, we can optimize
+
+            # simplest case: start and stop are positive
+            if start >= 0 and stop > 0:
+                if step < 0:
+                    start, stop = stop + 1, start + 1
+                return True, start, stop - start, False, slice(None, None, step if step != 1 else None)
+
+            # more complicated case: start and stop are negative
+            if start < 0 and stop < 0 and can_reverse:
+                if step < 0:
+                    start, stop = stop + 1, start + 1
+                step = -step
+                return True, - stop, stop - start, True, slice(None, None, step if step != 1 else None)
+
+        # for positive step, if we have a positive start, and no end or negative one, we have to retrieve all from
+        # the start and slice here
+        if step > 0 and (start is not None and start >= 0) and (stop is None or stop < 0):
+            return True, start, None, False, slice(None, stop, step)
+
+        # the reverse for negative step
+        if step < 0 and (start is None or start < 0) and (stop is not None and stop >= 0):
+            return True, stop + 1, None, False, slice(start, None, step)
+
+        # in all other cases we recover the whole collection and slice it in python
+        return False, None, None, False, slice(start, stop, step)
 
     def __getitem__(self, arg):
-        self._len_mode = False
-        self._slice = {}
-        if isinstance(arg, slice):
-            # A slice has been requested
-            # so add it to the sort parameters (via slice)
-            # and return the collection (a scliced collection is no more
-            # chainable, so we do not return `self`)
-            start = arg.start or 0
-            if start < 0:
-                # in case of a negative start, we can't use redis sort so
-                # we fetch all the collection before returning the wanted slice
-                return self._collection[arg]
-            self._slice['start'] = start
-            stop = arg.stop
-            # Redis expects a number of elements
-            # not a python style stop value
-            if stop is None:
-                # negative value for the count return all
-                self._slice['num'] = -1
+        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
+        old_sort = None if self._sort is None else self._sort.copy()
+        try:
+            self._len_mode = False
+            self._sort_limits = {}
+            self._optimized_slicing = False  # just to store the fact that the call was optimized
+            if isinstance(arg, slice):
+                # A slice has been requested
+                # We try to reduce the data to get back from redis
+                # when it's possible depending of the slice arguments.
+                # We use `self._sort_limits` to tell redis how to slice what
+                # it'll got from the `sort` command. We may also reverse
+                # `_sort['desc']`(the set will be read in reverse way as we wanted
+                # by redis).
+                # At the end we return the collection (a sliced collection
+                # is no more chainable, so we do not return `self`
+
+                optimized, start, count, rev, python_slice = self._optimize_slice(
+                    arg,
+                    can_reverse=True if not self._sort else self._sort.get('by', '') != 'nosort'
+                )
+
+                self._optimized_slicing = optimized
+
+                if optimized and start is None and count is None and python_slice == NONE_SLICE:
+                    return []
+
+                if start is not None or count is not None:
+                    self._sort_limits['start'] = start or 0
+                    self._sort_limits['num'] = -1 if count is None else count
+
+                if rev:
+                    if self._sort is None: self._sort = {}
+                    self._sort['desc'] = not self._sort.get('desc', False)
+
+                return self._collection[python_slice]
+
             else:
-                self._slice['num'] = stop - start
-            return self._collection
-        else:
-            # A single item has been requested
-            # Nevertheless, use the redis pagination, to minimize
-            # data transfert and use the fast redis offset system
-            start = arg
-            if start >= 0:
-                self._slice['start'] = start
-                self._slice['num'] = 1  # one element
-                return self._collection[0]
-            else:
-                # negative index, we have to fetch the whole collection first
-                return self._collection[start]
+                # A single item has been requested
+                # Nevertheless, use the redis pagination, to minimize
+                # data transfer and use the fast redis offset system
+                start = arg
+                self._sort_limits['num'] = 1  # one element
+                self._optimized_slicing = True
+                if start >= 0:
+                    self._sort_limits['start'] = start
+                    return self._collection[0]
+                else:
+                    # we sort the result in the reverse way, mark the final result as
+                    # reversed to re-reverse it at the end
+                    if self._sort is None: self._sort = {}
+                    self._sort['desc'] = not self._sort.get('desc', False)
+                    self._sort_limits['start'] = - start - 1
+                    return self._collection[0]
+        finally:
+            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode
+            self._sort = old_sort
 
     def _get_pk(self):
         """
@@ -100,13 +206,17 @@ class CollectionManager(object):
     def _prepare_sort_options(self, has_pk):
         """
         Prepare "sort" options to use when calling the collection, depending
-        on "_sort" and "_slice" attributes
+        on "_sort" and "_sort_limits" attributes
         """
         sort_options = {}
         if self._sort is not None and not has_pk:
             sort_options.update(self._sort)
-        if self._slice is not None:
-            sort_options.update(self._slice)
+        if self._sort_limits is not None:
+            if 'start' in self._sort_limits and 'num' not in self._sort_limits:
+                self._sort_limits['num'] = -1
+            elif 'num' in self._sort_limits and 'start' not in self._sort_limits:
+                self._sort_limits['start'] = 0
+            sort_options.update(self._sort_limits)
         if not sort_options and self._sort is None:
             sort_options = None
         return sort_options
@@ -116,7 +226,8 @@ class CollectionManager(object):
         """
         Effectively retrieve data according to lazy_collection.
         """
-        try:  # try block to always reset the _slice in the "finally" part
+        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
+        try:  # try block to always reset the _sort_limits in the "finally" part
 
             conn = self.cls.get_connection()
             self._len = 0
@@ -140,8 +251,12 @@ class CollectionManager(object):
 
             if self._len_mode:
                 if final_set is None:
-                    # final_set is None if we have a only pk
-                    self._len = 1
+                    if pk and not self._lazy_collection['sets']:
+                        # we have a only pk
+                        self._len = 1
+                    else:
+                        # we have nothing
+                        self._len = 0
                 else:
                     self._len = self._collection_length(final_set)
                     if keys_to_delete:
@@ -154,9 +269,13 @@ class CollectionManager(object):
 
                 # fill the collection
                 if final_set is None:
-                    # final_set is None if we have a pk without other sets, and no
-                    # needs to get values so we can simply return the pk
-                    collection = set([pk])
+                    if pk and not self._lazy_collection['sets']:
+                        # we have a pk without other sets, and no
+                        # needs to get values so we can simply return the pk
+                        collection = {pk}
+                    else:
+                        # we have nothing
+                        collection = {}
                 else:
                     # compute the sets and call redis te retrieve wanted values
                     collection = self._final_redis_call(final_set, sort_options)
@@ -165,18 +284,15 @@ class CollectionManager(object):
 
                 # Format return values if needed
                 return self._prepare_results(collection)
-
-        except:  # raise original exception
-            raise
-        finally:  # always reset the slice, having an exception or not
-            self._slice = {}
-            self._len_mode = True
+        finally:
+            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode
 
     def _final_redis_call(self, final_set, sort_options):
         """
         The final redis call to obtain the values to return from the "final_set"
         with some sort options.
         """
+
         conn = self.cls.get_connection()
         if sort_options is not None:
             # a sort, or values, call the SORT command on the set
@@ -224,7 +340,31 @@ class CollectionManager(object):
         Must return a tuple with a set of redis set keys, and another with
         new temporary keys to drop at the end of _get_final_set
         """
-        return (sets, set())
+
+        final_sets = set()
+        tmp_keys = set()
+
+        for set_ in sets:
+            if isinstance(set_, str):
+                final_sets.add(set_)
+            elif isinstance(set_, ParsedFilter):
+
+                for index_key, key_type, is_tmp in set_.index.get_filtered_keys(
+                            set_.suffix,
+                            accepted_key_types=self._accepted_key_types,
+                            *(set_.extra_field_parts + [set_.value])
+                        ):
+                    if key_type not in self._accepted_key_types:
+                        raise ValueError('The index key returned by the index %s is not valid' % (
+                            set_.index.__class__.__name__
+                        ))
+                    final_sets.add(index_key)
+                    if is_tmp:
+                        tmp_keys.add(index_key)
+            else:
+                raise ValueError('Invalid filter type')
+
+        return final_sets, tmp_keys
 
     def _get_final_set(self, sets, pk, sort_options):
         """
@@ -257,7 +397,10 @@ class CollectionManager(object):
             # no sets or pk, use the whole collection instead
             all_sets.add(self.cls.get_field('pk').collection_key)
 
-        if len(all_sets) == 1:
+        if not all_sets:
+            delete_set_later = False
+            final_set = None
+        elif len(all_sets) == 1:
             # if we have only one set, we  delete the set after calling
             # collection only if it's a temporary one, and we do not delete
             # it right now
@@ -290,20 +433,70 @@ class CollectionManager(object):
     def __call__(self, **filters):
         return self._add_filters(**filters)
 
+    def _field_is_pk(self, field_name):
+        """Check if the given name is the pk field, suffixed or not with "__eq" """
+        if self.cls._field_is_pk(field_name):
+            return True
+        if field_name.endswith('__eq') and self.cls._field_is_pk(field_name[:-4]):
+            return True
+        return False
+
+    def _parse_filter_key(self, key):
+        # Each key can have optional subpath
+        # We pass it as args to the field, which is responsable
+        # from handling them
+        # We only manage here the suffix handled by a filter
+
+        key_path = key.split('__')
+        field_name = key_path.pop(0)
+        field = self.cls.get_field(field_name)
+
+        if not field.indexable:
+            raise ImplementationError(
+                'Field %s.%s is not indexable' % (
+                    field._model.__name__, field.name
+                )
+            )
+
+        other_field_parts = key_path[:field._field_parts - 1]
+
+        if len(other_field_parts) + 1 != field._field_parts:
+            raise ImplementationError(
+                'Unexpected number of parts in filter %s for field %s.%s' % (
+                    key, field._model.__name__, field.name
+                )
+            )
+
+        rest = key_path[field._field_parts - 1:]
+        index_suffix = None if not rest else '__'.join(rest)
+        index_to_use = None
+        for index in field._indexes:
+            if index.can_handle_suffix(index_suffix):
+                index_to_use = index
+                break
+
+        if not index_to_use:
+            raise ImplementationError(
+                'No index found to manage filter %s for field %s.%s' % (
+                    key, field._model.__name__, field.name
+                )
+            )
+
+        return index_to_use, index_suffix, other_field_parts
+
     def _add_filters(self, **filters):
         """Define self._lazy_collection according to filters."""
         for key, value in filters.items():
-            if self.cls._field_is_pk(key):
+            if self._field_is_pk(key):
                 pk = self.cls.get_field('pk').normalize(value)
                 self._lazy_collection['pks'].add(pk)
             else:
-                # each key can have optional subpath
-                # we pass it as args to the field, which is responsable
-                # from handling them
-                key_path = key.split('__')
-                field_name = key_path.pop(0)
-                field = self.cls.get_field(field_name)
-                self._lazy_collection['sets'].add(field.index_key(value, *key_path))
+                # store the info to call the index later, in ``_prepare_sets``
+                # (to avoid doing extra work if the collection is never called)
+                index, suffix, extra_field_parts = self._parse_filter_key(key)
+                parsed_filter = ParsedFilter(index, suffix, extra_field_parts, value)
+                self._lazy_collection['sets'].append(parsed_filter)
+
         return self
 
     def __len__(self):
@@ -315,8 +508,12 @@ class CollectionManager(object):
         return self._len
 
     def __repr__(self):
-        self._len_mode = False
-        return self._collection.__repr__()
+        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
+        try:
+            self._len_mode = False
+            return self._collection.__repr__()
+        finally:
+            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode
 
     def instances(self, skip_exist_test=False):
         """
@@ -365,7 +562,10 @@ class CollectionManager(object):
             if by.startswith('-'):
                 parameters['desc'] = True
                 by = by[1:]
-            if self.cls.has_field(by):
+            if self.cls._field_is_pk(by):
+                # don't use field, the final set, which contains pks, will be sorted directly
+                del parameters['by']
+            elif self.cls.has_field(by):
                 # if we have a field, use its redis wildcard form
                 field = self.cls.get_field(by)
                 parameters['by'] = field.sort_wildcard
@@ -376,7 +576,7 @@ class CollectionManager(object):
         Parameters:
         `by`: pass either a field name or a wildcard string to sort on
               prefix with `-` to make a desc sort.
-        `alpha`: set it to True to sort lexicographilcally instead of numerically.
+        `alpha`: set it to True to sort lexicographically instead of numerically.
         """
         parameters = self._coerce_by_parameter(parameters)
         self._sort = parameters
