@@ -5,6 +5,7 @@ from future.utils import with_metaclass
 
 from logging import getLogger
 from copy import copy
+import inspect
 import threading
 
 from limpyd.fields import *
@@ -35,7 +36,36 @@ class MetaRedisModel(MetaRedisProxy):
             if not getattr(it, 'namespace', None):
                 it.namespace = ''
             it._name = ':'.join((it.namespace, name.lower()))
-            it.database._add_model(it)
+
+            # Get the caller (parent of the call of the `__new__`` method of the metaclass, which
+            # may be the current metaclass, or a one that inherit from it), with it's source file
+            # and line number
+            # It will be used in ``database._add_model``, if a model with the same name
+            # already exists, to check if it's the same one.
+            parent_frame = inspect.currentframe().f_back
+            while True:
+                f_code = parent_frame.f_code
+                f_locals = parent_frame.f_locals
+                if f_code.co_name == '__new__':
+                    try:
+                        # handle direct metaclass usage
+                        if issubclass(f_locals['mcs'], MetaRedisModel):
+                            parent_frame = parent_frame.f_back
+                            continue
+                    except Exception:
+                        pass
+                    # handle `future.utils.with_metaclass` and `six.with_metaclass`
+                    parent_cls = str(f_locals.get('cls'))
+                    if 'metaclass' in parent_cls and ('six.' in parent_cls or 'future.' in parent_cls):
+                        parent_frame = parent_frame.f_back
+                        continue
+                break
+            it._creation_source = (parent_frame.f_code.co_filename, parent_frame.f_lineno)
+            it_in_db = it.database._add_model(it)
+            # If the returned model is not the same, it's the one from the database
+            # we already added for this name, so we use it
+            if it_in_db is not it:
+                return it_in_db
 
         # init (or get from parents) lists of redis fields
         _fields = list(it._fields) if hasattr(it, '_fields') else []
@@ -122,6 +152,7 @@ class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
     abstract = True
     collection_manager = CollectionManager
     DoesNotExist = DoesNotExist
+    default_indexes = None
 
     available_getters = ('hmget', 'hgetall', 'hkeys', 'hvals', 'hlen')
     available_modifiers = ('hmset', 'hdel')
@@ -188,9 +219,8 @@ class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
                     raise ValueError(u"`%s` is not a valid field name "
                                       "for `%s`." % (field_name, self.__class__.__name__))
                 field = self.get_field(field_name)
-                if field.unique and self.exists(**{field_name: value}):
-                    raise UniquenessError(u"Field `%s` must be unique. "
-                                           "Value `%s` yet indexed." % (field.name, value))
+                if field.unique:
+                    field.check_uniqueness(value)
                 self._init_fields.add(field_name)
 
             # Do instanciate, starting by the pk and respecting fields order
@@ -207,6 +237,12 @@ class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
             self._pk = self.pk.normalize(args[0])
             self.connect()
 
+    @classmethod
+    def get_default_indexes(cls):
+        if cls.default_indexes is not None:
+            return cls.default_indexes
+        return cls.database.get_default_indexes()
+
     def connect(self):
         """
         Connect the instance to redis by checking the existence of its primary
@@ -311,7 +347,6 @@ class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
             field = self.get_field(field_name)
             if hasattr(field, "default"):
                 field.proxy_set(field.default)
-        delattr(self, '_init_fields')
 
     @classmethod
     def collection(cls, manager=None, **filters):
@@ -375,7 +410,12 @@ class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
                 elif len(result) > 1:
                     raise ValueError(u"More than one object matching filter: %s" % kwargs)
                 else:
-                    pk = result[0]
+                    try:
+                        pk = result[0]
+                    except IndexError:
+                        # object was deleted between the `len` check and now
+                        raise DoesNotExist(u"No object matching filter: %s" % kwargs)
+
         else:
             raise ValueError("Invalid `get` usage with args %s and kwargs %s" % (args, kwargs))
         return cls(pk)
@@ -389,8 +429,21 @@ class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
             inst = cls.get(**kwargs)
             created = False
         except DoesNotExist:
-            inst = cls(**kwargs)
-            created = True
+            try:
+                inst = cls(**kwargs)
+                created = True
+            except UniquenessError:
+                # This can happen if two ``get_or_connect`` where called in a very very short timespan:
+                # - Call 1 checks for existence
+                # - It does not exist and will create one
+                # - At this moment call 2 checks for existence
+                # - It does not exist and will create one
+                # - At this moment call 1 create the redis instance
+                # - But when call 2 wants to create the instance, it exists
+                # To solve this, in case of `UniquenessError`, it means it's already created so we
+                # can just get it from redis
+                inst = cls.get(**kwargs)
+                created = False
         except Exception:
             raise
         return inst, created
@@ -460,11 +513,11 @@ class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
             # We revert indexes previously set if we have an exception, then
             # really raise the error
             for field in indexed:
-                field._rollback_index()
+                field._rollback_indexes()
             raise
         finally:
             for field in indexed:
-                field._reset_index_cache()
+                field._reset_indexes_caches()
 
     def hdel(self, *args):
         """
@@ -524,3 +577,51 @@ class RedisModel(with_metaclass(MetaRedisModel, RedisProxyCommand)):
     @classmethod
     def _is_field_locked(cls, field):
         return field.name in cls._thread_lock_storage()
+
+    def scan_keys(self, count=None):
+        """Iter on all the key related to the current instance fields, using redis SCAN command
+
+        Parameters
+        ----------
+        count: int, default to None (redis uses 10)
+            Hint for redis about the number of expected result
+
+        Yields
+        -------
+        str
+            All keys found by the scan, one by one. A key can be returned multiple times, it's
+            related to the way the SCAN command works in redis.
+
+        """
+
+        pattern = self.make_key(
+            self._name,
+            self.pk.get(),
+            '*'
+        )
+
+        return self.database.scan_keys(pattern, count)
+
+    @classmethod
+    def scan_model_keys(cls, count=None):
+        """Iter on all the key related to the current model, using redis SCAN command
+
+        Parameters
+        ----------
+        count: int, default to None (redis uses 10)
+            Hint for redis about the number of expected result
+
+        Yields
+        -------
+        str
+            All keys found by the scan, one by one. A key can be returned multiple times, it's
+            related to the way the SCAN command works in redis.
+
+        """
+
+        pattern = cls.make_key(
+            cls._name,
+            "*",
+        )
+
+        return cls.database.scan_keys(pattern, count)
